@@ -1,5 +1,5 @@
 """Shared exception. Here because I want to have one root exception type"""
 
 
-class PIMSException(Exception):
+class PIMSError(Exception):
     pass
